//
// OpenGL Solid Modelling
//
// Nigel Stewart (nigels@eisa.net.au)
// RMIT Department of Manufacturing Systems Engineering
// 25th September 1998
//

#include <plane.h>
#include <math.h>

Plane::Plane()
: _d(0.0)
{
}

Plane::Plane(const Vector &dir,const double d)
: Vector(dir), _d(d)
{
}

Plane::Plane(const Vector &pnt,const Vector &dir)
: Vector(dir), _d(0.0)
{
	_d = -(pnt * dir);
}

int 
Plane::inside(const Vector &pnt) const
{
	return pnt * (*this) + _d >= 0.0;
}

double 
Plane::insideDist(const Vector &pnt) const
{
	return pnt * (*this) + _d;
}

Vector 
Plane::intersectRay(const Vector &a,const Vector &b) const
{
	double ad = insideDist(a);
	double bd = insideDist(b);
	
	double delta = fabs(ad/(bd-ad));

	return a + (b-a)*delta;
}
