//
// Demonstration of glCopyPixels roundoff problem.
//
// System Setup:
//	Windows NT 3.51, Pentium 100, 32Meg RAM
//	3Demon SX88, Display Driver 2.0, OpenGL driver 1.0.9
//	Visual C++ V4.0
//
// Strategy:
//	Draw red aux sphere in lower left quadrant of window.
//	Copy depth pixels to top right quadrant with glDepthFunc(GL_NOTEQUAL)
//  Draw red aux sphere in lower left quadrant with glDepthFunc(GL_EQUAL).
//	
// Results:
//  Expected result would be blue circle in bottom left & red top right
//  Roundoff causes some pixels to match (red) others to fail (blue).
//
// Compiling:
//		Required libraries: opengl32.lib glu32.lib glaux.lib
//
// Prepared by Nigel Stewart, RMIT University, 10-01-1999
// Email: nigels@eisa.net.au
// 
//

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glut.h>
#include <iostream.h>
#include <stdlib.h>

const GLubyte Red[3] =  { 255,  0,   0};
const GLubyte Blue[3] = {   0,  0, 255};

// #define DP glutSolidSphere(1.0,16,16);

#define zNear -1.2
#define zFar   -(zNear)

#define DP \
	glBegin(GL_POLYGON); \
	glVertex3i(-1,-1,zNear); \
	glVertex3i( 1,-1,zNear); \
	glVertex3i( 1, 1,zFar); \
	glVertex3i(-1, 1,zFar); \
	glEnd();
void Draw()
{
	GLuint viewport[4];
	GLuint width,height;

	glGetIntegerv(GL_VIEWPORT,(GLint *) viewport);
	width = viewport[2];
	height = viewport[3];

	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

	glViewport(0,0,width>>1,height);
	glColor3ubv(Blue);
	glDepthFunc(GL_LESS);
#if 0
	glRotatef(1.0,0.0,1.0,0.0);
	glRotatef(0.5,1.0,1.0,0.0);
#endif
	glPushMatrix();
	glPopMatrix();
	DP;

	glColor3ubv(Red);
	glDepthFunc(GL_EQUAL);
	DP;

/*
	glColor3ubv(Red);
	glDepthFunc(GL_EQUAL);
	glPushMatrix();
	glLoadIdentity();
	glRasterPos2d(-1,-1);
	glPopMatrix();
	glCopyPixels(0,0,width>>1,height,GL_DEPTH);
*/
	glViewport(0,0,width,height);
	glColor3ubv(Blue);
	glDepthFunc(GL_ALWAYS); // GL_NOTEQUAL);
	glPushMatrix();
	glLoadIdentity();
	glRasterPos2d(0,-1);
	glPopMatrix();
    glColor3ubv(Blue);
	glCopyPixels(0,0,width>>1,height,GL_DEPTH);
	
	glViewport(width>>1,0,width>>1,height);
	glColor3ubv(Red);
	glDepthFunc(GL_EQUAL);
	DP;
	
glViewport(0,0,width,height);

	glFinish();
	glutSwapBuffers(); 
	glutPostRedisplay();
}

void Reshape(int width, int height)
{
	glViewport(0, 0, width, height);
	cout << "Resize: " << width << "x" << height << endl;
}

void Key(unsigned char key,int x,int y)
{
	exit(0);
}

int main(int argc,char *argv[])
{
    glutInit(&argc,argv);
    glutInitWindowSize(500,250);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutCreateWindow("glCopyPixel Accuracy");

	glEnable(GL_DEPTH_TEST);
	glClearDepth(1.0);
	glDepthRange(0.0,1.0);

	glutKeyboardFunc(Key);
	glutReshapeFunc(Reshape);
	glutDisplayFunc(Draw);

	cout << "          Vendor : " << (char *) glGetString(GL_VENDOR) << endl;
	cout << "        Renderer : " << (char *) glGetString(GL_RENDERER) << endl;
	cout << "         Version : " << (char *) glGetString(GL_VERSION) << endl;

	GLint zDepth;
	GLint stencilDepth;
	GLint doubleBuffering;
	GLint rgbaMode;

	glGetIntegerv(GL_DEPTH_BITS,&zDepth);
	glGetIntegerv(GL_STENCIL_BITS,&stencilDepth);
	glGetIntegerv(GL_DOUBLEBUFFER,&doubleBuffering);
	glGetIntegerv(GL_RGBA_MODE,&rgbaMode);


	cout << "Double-buffering : " << (doubleBuffering ? "" : "not " ) << "supported." << endl;
	cout << "       RGBA mode : " << (rgbaMode ? "" : "not " ) << "enabled." << endl;
	cout << "    Depth buffer : " << zDepth << " bit-planes." << endl;
	cout << "  Stencil buffer : " << stencilDepth << " bit-planes." << endl;

	glutMainLoop();

	return 0;
}

