#include "TestString.h"
#include "GeneticAlg.h"

#include <math.h>
#include <stdio.h>
#include <assert.h>

#include <iomanip>
#include <map>

#include "GASimple.h"
#include "GA1DArra.h" 

typedef GA1DArrayAlleleGenome<ulong> OStringGenome;

OString toOString(ulong base,const OStringGenome &g)
{
   OString str(base);
	  
   for (uint o=0; o<g.length(); o++)
   {
      assert(g[o]<base);
      str.push_back(g[o]);
   }

   return str;
}

StringTest *ObjectStringTest = NULL;
float Objective(GAGenome &);

void geneticAlgorithm
(
	const uint n,
	const uint k,
	const uint l,
	const uint populationSize,
	StringTest &test,
	const uint seed
)
{
	if (seed)
		GARandomSeed(seed);

	// Configure objective function context

	ObjectStringTest = &test;
	assert(ObjectStringTest);

	// Create first individual

	OStringGenome genome(l,GAAlleleSet<ulong>(0,n),Objective);   

	// Set the resize behaviour

	genome.resizeBehaviour(l,l);		

	GASimpleGA ga(genome);
	ga.populationSize(populationSize);
	ga.nGenerations(1);
	ga.elitist(gaTrue);
	ga.evolve();

	do
	{
		ga.step();

		OStringGenome *g = ((OStringGenome *) &ga.statistics().bestIndividual());
		cout << setw(5) << ga.statistics().generation() << ' ';
		cout.setf(ios::fixed);
		cout << setprecision(5) << ga.statistics().maxEver() << ' ';
		cout << toOString(n,*g) << endl;
	}
	while (ga.statistics().maxEver()<1.0);

	cout << endl;
	cout << ga.statistics() << endl;

	{
		OStringGenome *g = ((OStringGenome *) &ga.statistics().bestIndividual());
		test.accepted().push_back(toOString(n,*g));
	}

	return;
}
 
float
Objective(GAGenome &g) 
{
  assert(ObjectStringTest);
  StringTest &test = *ObjectStringTest;

  test.testCount()++;
  test.testCountTotal()++;

  OStringGenome &genome = (OStringGenome &) g;

  assert(test.criteria().size());

  OString str = toOString(test.criteria()[0].base(),genome);

  ulong match = 0;

  for (ulong p=0; p<test.criteria().size(); p++)
    if (str.insideByDecimation(test.criteria()[p]))
      match++;

  return (double) match / (double) test.criteria().size();
}



