#include "TestString.h"
#include "altGroup.h"

#include <algorithm>

/*

vector<OString> testAgainst;
vector<OString> testAccept;
ulong           testStringCount = 0;
bool testNormalise = true;

//
// Test string for fitness, print it to std out
// if it satisfies all permutation in perms
//

void _testStringAllByDecimation(const OString &str,const vector<OString> &against)
{
	testStringCount++;

	if (!against.size())
		return;

	ulong match = 0;

	for (ulong p=0; p<against.size(); p++)
		if (!str.insideByDecimation(against[p]))
			break;
		else
			match++;
	
	if (match==against.size())
	{	
		OString strNorm(testNormalise ? str.normalise() : str);

		if (find(testAccept.begin(),testAccept.end(),strNorm)==testAccept.end())
		{
			testAccept.push_back(strNorm);
			displayString(strNorm);
		}
	}
}

void testStringAllByDecimation(const OString &str)
{
	_testStringAllByDecimation(str,testAgainst);
}

//
//
//

void _testStringNoneByDecimation(const OString &str,const vector<OString> &against)
{
	testStringCount++;

	if (!against.size())
		return;

	ulong match = 0;

	for (ulong p=0; p<against.size(); p++)
		if (str.insideByDecimation(against[p]))
			return;

	testAccept.push_back(str);
	displayString(str);
}

void testStringNoneByDecimation(const OString &str)
{
	_testStringNoneByDecimation(str,testAgainst);
}

*/

ulong StringTest::_testCountTotal = 0;

StringTest::StringTest(const StringTestType testType,const bool histogram)
: _testType(testType), _testCount(0), _enableHistogram(histogram)
{
}

vector<OString> &
StringTest::criteria()
{
	return _criteria;
}

vector<OString> &
StringTest::accepted()
{
	return _accepted;
}

int 
StringTest::test(const OString &str)
{
	_testCount++;
	_testCountTotal++;

	if (!_criteria.size())
		return 0;

	switch (_testType)
	{
		case STRING_TEST_ALL_BY_DECIMATION:
		{
			ulong match = 0;

			for (ulong p=0; p<_criteria.size(); p++)
				if (!str.insideByDecimation(_criteria[p]))
				{
					if (!_enableHistogram)
						break;
				}
				else
					match++;
			
			if (_enableHistogram)
			{
				if (!_histogram.size())
				{
					_histogram.clear();
					_histogram.resize(_criteria.size()+1,0);
				}

				_histogram[match]++;
			}

			if (match==_criteria.size())
			{
				addAccepted(str);
				return 1;
			}
		}
		break;

		case STRING_TEST_NONE_BY_DECIMATION:
		{
			for (ulong p=0; p<_criteria.size(); p++)
				if (str.insideByDecimation(_criteria[p]))
					return 0;

			addAccepted(str);
			return 1;
		}
		break;
	}

	return 0;
}

void
StringTest::histogram(ostream &os)
{
	if (_histogram.size())
		for (ulong f=0; f<_histogram.size(); f++)
			cout << f << ' ' << _histogram[f] << endl;
}

void 
StringTest::addAccepted(const OString &str)
{
	if (_moreCriteria.size())
	{
		for (int c=0; c<_moreCriteria.size(); c++)
			if (!_moreCriteria[c].test(str))
				return;
	}

	if (find(_accepted.begin(),_accepted.end(),str)==_accepted.end())
	{
		_accepted.push_back(str);
		displayString(str);
	}
}

ulong &
StringTest::testCount()
{
	return _testCount;
}

ulong &
StringTest::testCountTotal()
{
	return _testCountTotal;
}

vector<StringTest> &
StringTest::moreCriteria()
{
	return _moreCriteria;
}

void 
StringTest::reset()
{
	_testCount = 0;
	_histogram.clear();
}

/*

private:

  StringTestType _testType;

  vector<OString> _criteria;
  vector<OString> _accepted;

         ulong _testCount;
  static ulong _testCountTotal;

  void testStringAllByDecimation(const OString &str);
  void testStringNoneByDecimation(const OString &str);

};
*/