#include "NikGen.h"

//
// Recursive generation
//
// r     remaining recursions until ready for test
// base  the generated string "so far"
// perms the set of permutations of (n-1) characters
// test  function to test the fitness of each generated string
//

static void 
nikLygerosGen
(
	const int r,
	OString base,
	const vector<OString> &perms,
	void (*test)(const OString &)
)
{
	//
	// Test the string
	//

	if (r==0)
	{
		base.push_back(0);
#ifdef _DEBUG
		cout << "Testing: " << base << endl;
#endif
		test(base);
	}
	else

		//
		// Append to the string base
		//

		for (uint p=0; p<perms.size(); p++)
		{
			OString next = base;		// Copy the base
			next.push_back(0);			// insert an 'a'

			// All but the first+last recursion uses only
			// the first (n-2) characters of each
			// permutation of (n-1).

			if (r>1)
				for (uint i=0; i<perms[p].size()-1; i++)
					next.push_back((perms[p])[i]);
			else
				next += perms[p];

			// Recurse

			nikLygerosGen(r-1,next,perms,test);
		}
}

// Main loop for Nik Lygeros generation

void 
nikLygeros(const ulong n,void (*test)(const OString &))
{
	if (n<1)
		return;

	// Generate all permutations of bcdef...

	vector<OString> perms = OString::permutations(n-1);

	// Adjust from abcde.. to bcdef..

	for (uint i=0; i<perms.size(); i++)
	{
		perms[i] += 1;
#ifdef _DEBUG
		cout << perms[i] << endl;
#endif
	}

	// String always begins with abcdef

	OString base(n);

	for (i=0; i<n; i++)
		base.push_back(i);

#ifdef _DEBUG
	cout << "Base is " << base << endl;
	cout << "There are " << perms.size() << " permutations of n-1." << endl;
#endif

	nikLygerosGen(n-2,base,perms,test);
}
