#ifndef _OString_h
#define _OString_h

#include <vector>
#include <string>
#include <iostream>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

using namespace std;

typedef unsigned long ulong;
typedef unsigned int  uint;

//

ulong factorial(const ulong n);
ulong power(const ulong a,const ulong b);

//

class OString : public vector<ulong>
{
public:
	OString(const ulong n = 2,const ulong l = 0);
	OString(const char *str);
	OString(std::istream &is);

	ulong base() const;

	// IO

	istream &read(istream &is);
	ostream &write(ostream &os) const;

	static ostream &write(ostream &os,const vector<OString> &str);
	static istream &read(istream &is,vector<OString> &str);

	friend ostream &operator<<(ostream &os,const OString &str);
	friend ostream &operator<<(ostream &os,vector<OString> &str);

	// Lexicographic neighbours

	bool operator++();
	bool operator--();

	// Comparison

	bool operator==(const OString &str) const;
	bool operator!=(const OString &str) const;
	bool operator==(const vector<OString> &str) const;
	bool operator!=(const vector<OString> &str) const;

	// Catenate

	OString  operator+ (const OString &b);
	OString &operator+=(const OString &b);

	// Offset each element

	OString  operator+ (const ulong &x);
	OString &operator+=(const ulong &x);

	//
	// Is another string contained by 'decimation'?
	//

	bool    insideByDecimation(const OString &str) const;

	//
	// Swap two entries in the string
	//

	OString transposition(const uint a,const uint b) const;

	//
	// Remap entries according to provided
	// string.  'a' becomes first entry,
	// 'b' the second, and so on...
	//
	
	OString substitute(const OString s) const;

	//
	// Substitute such that 'a' is first unqiue, 'b' is second, etc..
	//

	OString normalise() const;

	//
	// Mirror the string such that 'abcde' -> 'edcba'
	//

	OString mirror() const;

	//
	// Extract all but n'th character in string
	//

	OString skipNth(const uint pos) const;

	//
	
	static OString              random(const uint n,const uint l);
	static OString              randomNoDup(const uint n,const uint l);
	static OString              permutation(const ulong n,const ulong p,const ulong k=0);
	static vector<OString>		permutations(const ulong n,const ulong k=0);

	// Comparison

	// Conversion to String

	operator string() const;

	static bool asAlpha;

private:

	typedef vector<ulong>::iterator iterator;

	ulong _n;
};

extern void displayString(const OString &str);

#endif