//
// OpenGL Solid Modelling
//
// Nigel Stewart (nigels@eisa.net.au)
// RMIT Department of Mechanical and Manufacturing Engineering
// 11th December 1997
//

#ifndef ATYPES_H
#define ATYPES_H

#ifdef WIN32
#include <bool.h>
#else
#include <stl_config.h>  // SGI 
#endif
#include <float.h>

//
// Basic types, constants
// and functions
//

typedef double			Real;
typedef Real			Angle;
typedef unsigned int	Uint;

const Real NaN = FLT_MAX;

#ifndef Pi
const Real Pi=3.14159265358979323846;
#endif
const Real Degr=Pi/180.0;
const Real Pi2=2*Pi;

template <class A>				A SQ(A a) { return a*a; }
template <class A, class B>		A SIGN(A a, B b) { return b<0?  -a: a; }
template <class A>				bool SAME_SIGN(A a, A b) { return (a<0 && b<0   ||   a>0 && b>0); }
template <class A>				A CLAMP(const A x, const A min_v, const A max_v) { return x<min_v?  min_v: (x>max_v? max_v: x); }

#ifdef WIN32
#define MAX(x,y) (((x)>(y)) ? (x) : (y))
#define MIN(x,y) (((x)<(y)) ? (x) : (y))
#endif

#endif

