//
// OpenGL Solid Modelling
//
// Nigel Stewart (nigels@eisa.net.au)
// RMIT Department of Mechanical and Manufacturing Engineering
// 3rd January 1999
//

#ifndef CSGPRIM_H
#define CSGPRIM_H

#include <map.h>
#include <csgtree.h>
#include <iostream.h>

#ifdef WIN32
#pragma warning(disable : 4786)     // Symbol truncation for debugger 
#endif

template <class T>
struct StrCmp : binary_function<T, T, bool> {
    bool operator()(const T& x, const T& y) const { return strcmp(x,y)<0; }
};

class CsgPrimitiveDatabase
{
public:
	CsgPrimitiveDatabase();

	bool   add(char *,const GLuint name);
	bool   add(char *,CsgTreeLeaf *leaf);

	GLuint       lookupDL(char *);
	CsgTreeLeaf *lookupP(char *);

	ostream &writePOV(ostream &os,const CsgTreeLeaf &leaf);
	ostream &writePOV(ostream &os,const CsgTreeNode &tree);

private:

	map<char *,CsgTreeLeaf *,StrCmp<char *> > _namedPrimitives;
	map<char *,GLuint,StrCmp<char *> >        _displayLists;
};

#endif

