#include <func.h>

#include <math.h>
#include <assert.h>

#define SQR(x) ((x)*(x))

Line::Line(const Vector origin,const Vector direction)
: _origin(origin), _direction(direction)
{
}

Vector
Line::f(double t) const
{
	return _origin + _direction*t;
}

Vector 
Line::df(double t) const
{
	return _direction;
}

Vector 
Line::ddf(double t) const
{
	return Vector0;
}


Helix::Helix(const Vector origin,const Vector direction,const double radius,const double freq)
: _origin(origin), _direction(direction), _radius(radius), _freq(freq)
{
	Vector dir = _direction;
	dir.normalise();
	double angle = acos(dir*VectorX);

	if (fabs(angle)>0.01)
		_trans = matrixRotate(xProduct(_direction,VectorX),angle);
}

Vector
Helix::f(double t) const
{
	Vector orbit = VectorY*sin(t*_freq*Pi2) + VectorZ*cos(t*_freq*Pi2);
	orbit = orbit * _radius;
	return _origin + _direction*t + _trans*orbit; 
}

Vector 
Helix::df(double t) const
{
	Vector orbit = VectorY*cos(t*_freq*Pi2) - VectorZ*sin(t*_freq*Pi2);
	orbit = orbit * _radius * _freq * Pi2;

	return _direction + _trans*orbit;
}

Vector 
Helix::ddf(double t) const
{
	Vector orbit = -VectorY*sin(t*_freq*Pi2) - VectorZ*cos(t*_freq*Pi2);
	orbit = orbit * _radius * SQR(_freq*Pi2);

	return _trans*orbit;
}

//
//
//

FuncSegment::FuncSegment(const Func &f,const double begin,const double end)
: _func(f), _begin(begin), _end(end)
{
	assert(_begin<_end);
	assert(CLAMP(_begin,0.0,1.0)==_begin);
	assert(CLAMP(_end,0.0,1.0)==_end);
}


Vector 
FuncSegment::f(double t) const
{
	return _func.f(segT(t));
}

Vector 
FuncSegment::df(double t) const
{
	return _func.df(segT(t));
}

Vector 
FuncSegment::ddf(double t) const
{
	return _func.ddf(segT(t));
}

double 
FuncSegment::segT(const double t) const
{
	return (_end-_begin)*t+_begin;
}



