#ifndef FUNC_H
#define FUNC_H

#include <vector3.h>
#include <matrix4.h>

class Func
{
public:
	virtual Vector f(double t) const = 0;		// f(t)
	virtual Vector df(double t) const = 0;	    // f'(t)
	virtual Vector ddf(double t) const = 0;	    // f''(t)
};

class Line : public Func
{
public:
	Line(const Vector origin,const Vector direction);

	Vector f(double t) const;
	Vector df(double t) const;
	Vector ddf(double t) const;

protected:
	Vector _origin;
	Vector _direction;
};

class Helix : public Func
{
public:
	Helix(const Vector origin,const Vector direction,const double radius,const double freq);

	Vector f(double t) const;
	Vector df(double t) const;
	Vector ddf(double t) const;

protected:
	Matrix _trans;
	Vector _origin;
	Vector _direction;
	double _radius;
	double _freq;
};

class FuncSegment : public Func
{
public:

	FuncSegment(const Func &f,const double begin,const double end);

	Vector f(double t) const;
	Vector df(double t) const;
	Vector ddf(double t) const;

protected:

	double segT(const double t) const;

	const Func &_func;
	double _begin,_end;
};

#endif

