#include <mesh.h>
#include <map.h>

#include <fstream.h>

#include <read3dsb.h>
#include <vrmlpars.h>

#include <stdio.h>
#include <string.h>

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>

Mesh *currentMesh = NULL;

void 
insertMesh(vector<Point> &vertex,vector<TriangleIndex> &triangle,char *name)
{
	if (currentMesh)
	{
		int base = currentMesh->point().size();

		for (Uint p=0;p<vertex.size();p++)
			currentMesh->point().push_back(vertex[p]);

		for (Uint t=0;t<triangle.size();t++)
			currentMesh->triangle().push_back
			(
				Triangle
				(
					&currentMesh->point()[base+triangle[t].a],
					&currentMesh->point()[base+triangle[t].b],
					&currentMesh->point()[base+triangle[t].c]
				)
			);
	}
}

Mesh::Mesh()
{
}

Mesh::Mesh(char *filename,MeshFileType type)
{
	if (type==MESH_AUTO)
	{
		char *ext = filename + strlen(filename) - 3;

		if (ext<filename)
			return;

		if ( !strcmp(ext,"3ds") ||  !strcmp(ext,"3DS") ) type = MESH_3DS_BINARY;
		if ( !strcmp(ext,"wrl") ||  !strcmp(ext,"WRL") ) type = MESH_VRML;
		if ( !strcmp(ext,"off") ||  !strcmp(ext,"OFF") ) type = MESH_OFF;
	}

	switch (type)
	{
	case MESH_3DS_BINARY:
		{
			fstream is(filename,ios::in | ios::nocreate );
#ifdef _SGIAPI
			if (!is.good())
#else
			if (!is.is_open())
#endif
				return;
			currentMesh = this;
			read3DSbinary(is,insertMesh);
			currentMesh = NULL;
		}
		break;
		
	case MESH_VRML:
		{
			yyVrmlMesh = this;
			Vrmlin = fopen(filename,"rt");

			if (Vrmlin)
			{
				Vrmlparse();
				fclose(Vrmlin);
			}

			Vrmlin = NULL;
			yyVrmlMesh = NULL;
		}
		break;

	case MESH_OFF:
		{

			fstream is(filename,ios::in | ios::nocreate);
#ifdef _SGIAPI
			if (!is.good())
#else
			if (!is.is_open())
#endif
				return;
			loadOff(is);
		}
		break;

	default:
		break;
	}
}

void 
Mesh::loadOff(istream &is)
{
	int dimensions;
	int vertex,faces,cells;

	is >> dimensions;

	if (dimensions!=2 && dimensions!=3)
		return;

	is >> vertex;
	is >> faces;
	is >> cells;

	for (int v=0; v<vertex; v++)
	{
		double x,y,z;

		is >> x;
		is >> y;

		if (dimensions==3)
			is >> z;
		else
			z = 0.0;

		_point.push_back( Point(Vector(x,y,z), Vector0) );
	}

	for (int f=0; f<faces; f++)
	{
		int number;

		is >> number;
		
		switch (dimensions)
		{
		case 2:
			{
			}
		case 3:
			{
				switch (number)
				{
				case 3:					// Triangle
					{
						int a,b,c;

						is >> a;
						is >> b;
						is >> c;

						_triangle.push_back(Triangle(&_point[a],&_point[b],&_point[c]));
					}
					break;
				case 4:						// Prism
					{
						int a,b,c,d;

						is >> a;
						is >> b;
						is >> c;
						is >> d;

						_triangle.push_back(Triangle(&_point[a],&_point[b],&_point[c]));
						_triangle.push_back(Triangle(&_point[a],&_point[b],&_point[d]));
						_triangle.push_back(Triangle(&_point[a],&_point[c],&_point[d]));
						_triangle.push_back(Triangle(&_point[b],&_point[c],&_point[d]));
					}
					break;
				}
			}
			break;
		}
	}
}

ostream &
Mesh::writePOV(ostream &os) 
{
	os << "mesh {" << endl;

		for (Uint t=0; t<_triangle.size(); t++)
			_triangle[t].writePOV(os);

	os << "}" << endl;

	return os;
}

ostream &
Mesh::writeAOff(ostream &os) 
{
	map<Point *,int,greater<void *> > mapPoint;

	os << _point.size() << ' ' << _triangle.size() << ' ' << 0 << endl;

	for (Uint v=0; v<_point.size(); v++)
	{
		mapPoint[&_point[v]] = v;

		os << _point[v][0] << ' ';
		os << _point[v][1] << ' ';
		os << _point[v][2] << endl;
	}

	for (Uint t=0;t<_triangle.size(); t++)
	{
		os << 3 << '\t';
		os << mapPoint[_triangle[t][0]] + 1 << '\t';
		os << mapPoint[_triangle[t][1]] + 1 << '\t';
		os << mapPoint[_triangle[t][2]] + 1 << endl;
	}

	return os;
}

void
Mesh::glDraw()
{
	glPushAttrib(GL_ENABLE_BIT);

//		glDisable(GL_CULL_FACE);
		glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);

		glBegin(GL_TRIANGLES);


		for (Uint t=0; t<_triangle.size(); t++)
			_triangle[t].glDraw();

		glEnd();

	glPopAttrib();
}

