#include <parsetr.h>

#include <colours.h>
#include <csgprim.h>
#include <vector.h>
#include <mesh.h>

#include <string.h>
#include <stdio.h>

#define EAT_WHITE(str) while (**(str)==' ' || **(str)=='\t') (*(str))++;
#define NEXT_WHITE(str) while (**(str)!='\0' && **(str)!=' ' && **(str)!='\t') (*(str))++;

CsgTreeNodeType parseOperator(char **buffer)
{
	char buf[1024];

	EAT_WHITE(buffer);
	sscanf(*buffer,"%s",buf);

	if (strlen(buf)!=1)
	{
		cout << "Expecting operator, ERROR : " << *buffer << endl;
		exit(0);
	}

	CsgTreeNodeType op;

	switch (*buf)
	{
		case '+': op = UNION_NODE; break;
		case '-': op = DIFFERENCE_NODE;	break;
		case '.': op = INTERSECTION_NODE; break;
		default:
			{
				cout << "Invalid operator, ERROR: " << *buffer << endl;
				exit(0);
			}
			break;
	}

	(*buffer)++;
	EAT_WHITE(buffer);

	return op;
}

CsgTreeNode *parseTree(CsgPrimitiveDatabase &db,char **buffer)
{
	char buf[1024];

	EAT_WHITE(buffer);

	if (**buffer=='(')
	{
		(*buffer)++;
		CsgTreeNode *tree = parseTree(db,buffer);

		EAT_WHITE(buffer);
		if ( **buffer!=')' )
		{
			cout << "Expecting closing backet, ERROR: " << *buffer << endl;
			exit(0);
		}

		(*buffer)++;
		EAT_WHITE(buffer);

		if ( **buffer=='\0' || **buffer==')' )
			return tree;

		CsgTreeNodeType op = parseOperator(buffer);
		
		CsgTreeNode *root = new CsgTreeNode(op);
		root->left() = tree;

		if (**buffer=='(')
			root->right() = parseTree(db,buffer);
		else
		{
			sscanf(*buffer,"%s",buf);
			root->right() = db.lookupP(buf);

			if (!root->right())
			{
				cout << "Expecting primitive, ERROR : " << *buffer << endl;
				exit(0);
			}

			*buffer += strlen(buf);
			EAT_WHITE(buffer);
		}

		return root;
	}
	else
	{
		CsgTreeNode *left = NULL;
		CsgTreeNode *right = NULL;

		EAT_WHITE(buffer);

		sscanf(*buffer,"%s",buf);

		CsgTreeLeaf *p = db.lookupP(buf);

		if (!p)
		{
			cout << "Expecting primitive, ERROR : " << *buffer << endl;
			exit(0);
		}

		*buffer += strlen(buf);

		CsgTreeNodeType op = parseOperator(buffer);
		
		CsgTreeNode *tree = new CsgTreeNode(op);
		tree->left() = p;

		if (**buffer=='(')
			tree->right() = parseTree(db,buffer);
		else
		{
			sscanf(*buffer,"%s",buf);
			tree->right() = db.lookupP(buf);

			if (!tree->right())
			{
				cout << "Expecting primitive, ERROR : " << *buffer << endl;
				exit(0);
			}

			*buffer += strlen(buf);
			EAT_WHITE(buffer);
		}

		return tree;
	}
}

double
parseDouble(char **buffer)
{
	EAT_WHITE(buffer);
	double v = atof(*buffer);
	NEXT_WHITE(buffer);
	EAT_WHITE(buffer);

	return v;
}

Vector
parseVector(char **buffer)
{
	EAT_WHITE(buffer);

	Vector v;

	v[0] = parseDouble(buffer);
	v[1] = parseDouble(buffer);
	v[2] = parseDouble(buffer);

	return v;
}

Matrix
parseMatrix(char **buffer)
{
	EAT_WHITE(buffer);

	if (!strncmp(*buffer,"scale",5))
	{
		(*buffer) += 5;
		EAT_WHITE(buffer);
		Vector scale = parseVector(buffer);
		return matrixScale(scale) * parseMatrix(buffer);
	}

	if (!strncmp(*buffer,"translate",9))
	{
		(*buffer) += 9;
		EAT_WHITE(buffer);
		Vector translate = parseVector(buffer); 
		return matrixTranslate(translate) * parseMatrix(buffer);
	}

	if (!strncmp(*buffer,"rotate",6))
	{
		(*buffer) += 6;
		EAT_WHITE(buffer);
		Vector axis = parseVector(buffer);
		double angle = parseDouble(buffer);
		return matrixRotate(axis,angle) * parseMatrix(buffer);
	}

	return Matrix();
}

  
vector<CsgTreeNode *> parseTrees(CsgPrimitiveDatabase &db,istream &is)
{
	//
	//
	//

	vector<CsgTreeNode *> trees;

	const int bufferSize = 2048;
	char buffer[2048];
	char *pos;

	while (1)
	{
		pos = buffer;

		if (is.eof())
			break;

		is.getline(pos,2048);

		EAT_WHITE(&pos);

		char id[1024];
		if (sscanf(pos,"%[^= \t\n]",id)!=1)
			continue;

		pos += strlen(id);
		EAT_WHITE(&pos);

		if (!strcmp("Tree",id) && *pos=='=')
		{
			pos++;
			trees.push_back(parseTree(db,&pos)); 

			cout << *trees.back() << endl;
		}
		else
		{
			char name[1024];
			char col[1024];

			if ( *pos!='=' )
			{
				cout << "Expecting '=' ERROR : " << pos << endl;
				exit(0);
			}

			pos++;
			EAT_WHITE(&pos);

			sscanf(pos,"%s",name);
			
			NEXT_WHITE(&pos);
			EAT_WHITE(&pos);

			sscanf(pos,"%s",col);

			NEXT_WHITE(&pos);

			cout << id << " = " << name << ' ' << col << endl;

			char *tmp = strdup(id);
			GLuint dl = db.lookupDL(name);

			if (!dl)
			{
				cout << "Opening " << name << "... "; cout.flush();

				Mesh mesh(name);

				if (mesh.point().size())
				{
					cout << "OK." << endl;

					dl = glGenLists(1);
					glNewList(dl,GL_COMPILE);
						mesh.glDraw();
					glEndList();
				}
			}
			
			CsgTreeLeaf *leaf = new CsgTreeLeaf(tmp,dl,colour(col),parseMatrix(&pos));
			db.add(tmp,leaf);
			trees.push_back(leaf);

			// REVISIT

			if (!strcmp(name,"helix"))
				leaf->convexity() = 3;
		}
	}

	return trees;
}

