#include <subtract.h>
#include <glbuffer.h>
#include <zplane.h>
#include <goldfeat.h>

#include <iostream.h>
#include <assert.h>
#include <string.h>
#include <math.h>

inline void glDepthColorStencilMasks(const GLboolean depth,const GLboolean color,const GLuint stencil)
{
	glDepthMask(depth); 
	glColorMask(color,color,color,color); 
	glStencilMask(stencil);
}

inline void glDepthStencilCullingEnable(const GLboolean depth,const GLboolean stencil,const GLboolean culling)
{
	if (depth)   
		glEnable(GL_DEPTH_TEST); 
	else 
		glDisable(GL_DEPTH_TEST); 

	if (stencil) 
		glEnable(GL_STENCIL_TEST); 
	else 
		glDisable(GL_STENCIL_TEST); 

	if (culling) 
		glEnable(GL_CULL_FACE);    
	else 
		glDisable(GL_CULL_FACE); 
}

GLuint mostSignificantStencilBit = (1<<7);

/////////////////////////////////////////////////////
// closestSurfaceBehind
//
//

void
closestSurfaceBehind(const CsgProduct &product,const vector<unsigned int> &convexity)
{
	assert(product.complemented().size()==convexity.size());

	#ifndef NDEBUG
	cout << "Entering closestSurfaceBehind..." << endl;
	#endif

	GLuint failBit = mostSignificantStencilBit;

	//
	// Decrement nonzero counts in stencil buffer
	//

	glDepthColorStencilMasks(GL_FALSE,GL_FALSE,0xffff);
	glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_FALSE);
	glDepthFunc(GL_ALWAYS);
	glStencilFunc(GL_NOTEQUAL,0,0xffff);
	glStencilOp(GL_KEEP,GL_DECR,GL_DECR);
	drawZnear();

	//
	// Save copy of stencil buffer
	//

	stencilBuffer stencil;

	//
	// Save copy of depth Buffer
	//

	depthBuffer   depth;

	//
	// Create scratch depth buffer
	//

	depthBuffer   scratch(NULL);
	scratch = depth.maxPossible();

	//
	// Inner Loop
	//

	for (unsigned int o=0; o<product.complemented().size(); o++)
	{
		for (GLuint k=0; k<convexity[o]; k++)
		{

			/* glPixelTransferf(GL_DEPTH_BIAS,depthBias); */
				
			glDepthColorStencilMasks(GL_TRUE,GL_FALSE,0x0000);
			glDepthStencilCullingEnable(GL_TRUE,GL_FALSE,GL_FALSE);
			glDepthFunc(GL_ALWAYS);
			depth.write();

			/* glPixelTransferf(GL_DEPTH_BIAS,0.0); */

			//
			// Initialise stencil with FAIL flag set to true
			//

			glDepthColorStencilMasks(GL_FALSE,GL_FALSE,0xffff);
			glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_FALSE);
			glDepthFunc(GL_ALWAYS);
			glStencilFunc(GL_ALWAYS,failBit,0xffff);
			glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);
			drawZnear();

			//
			//	If k-th layer greater than Depth, set FAIL bit to zero.
			//

			glDepthColorStencilMasks(GL_FALSE,GL_FALSE,0xffff); 
			glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_TRUE);
			glDepthFunc(GL_LEQUAL);
			glStencilFunc(GL_EQUAL, k | failBit, 0xffff);
			glStencilOp(GL_INCR,GL_ZERO,GL_INCR);
			glDisable(GL_LIGHTING);
			product.complemented()[o]->draw();

			//
			//  Clear count field
			//

			glDepthColorStencilMasks(GL_FALSE,GL_FALSE,0xffff ^ failBit);
			glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_FALSE);
			glDepthFunc(GL_ALWAYS);
			glStencilFunc(GL_ALWAYS,0x0000,0xffff);
			glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);
			drawZnear();

			//
			// Switch to scratchDepth
			//

			glDepthColorStencilMasks(GL_TRUE,GL_FALSE,0x0000);
			glDepthStencilCullingEnable(GL_TRUE,GL_FALSE,GL_FALSE);
			glDepthFunc(GL_ALWAYS);
			scratch.write();

			//
			// Store Z of surface closer than Depth
			//

			glDepthColorStencilMasks(GL_TRUE,GL_FALSE,0xffff); 
			glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_TRUE);
			glDepthFunc(GL_LESS);
			glStencilFunc(GL_EQUAL,k,0xffff);
			glStencilOp(GL_INCR,GL_INCR,GL_INCR);

			product.complemented()[o]->draw();
		
			//
			// Preserve scratchDepth
			//

			scratch.read();
		}
	}

	//
	// Restore Depth and Stencil
	//

	/* glPixelTransferf(GL_DEPTH_BIAS,depthBias); */

	glDepthColorStencilMasks(GL_TRUE,GL_FALSE,0x0000);
	glDepthStencilCullingEnable(GL_TRUE,GL_FALSE,GL_FALSE);
	glDepthFunc(GL_ALWAYS);
	depth.write();
	
	/* glPixelTransferf(GL_DEPTH_BIAS,0.0); */

	stencil.write();

	glDepthColorStencilMasks(GL_TRUE,GL_FALSE,failBit);
	glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_FALSE);		
	glDepthFunc(GL_ALWAYS);
	glStencilFunc(GL_EQUAL,0,0xffff);
	glStencilOp(GL_KEEP,GL_KEEP,GL_INVERT);

	scratch.write();

	//
	// For areas that EMPTY flag is set, count number of faces.
	//

	glDepthColorStencilMasks(GL_FALSE,GL_FALSE,0xffff ^ failBit);
	glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_TRUE);	
	glDepthFunc(GL_EQUAL);
	glStencilFunc(GL_EQUAL,failBit,failBit);
	glStencilOp(GL_KEEP,GL_KEEP,GL_INCR);
	for (Uint prim=0; prim<product.complemented().size(); prim++)
		product.complemented()[prim]->draw();

	//
	// Clear EMPTY flag
	//

	glDepthColorStencilMasks(GL_FALSE,GL_FALSE,failBit);
	glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_FALSE);	
	glDepthFunc(GL_ALWAYS);
	glStencilFunc(GL_ALWAYS,0x0000,0x0000);
	glStencilOp(GL_KEEP,GL_KEEP,GL_ZERO);
	drawZnear();

	//
	// Leave things sane.
	//

	glDepthColorStencilMasks(GL_TRUE,GL_TRUE,0xffff);
	glDepthStencilCullingEnable(GL_TRUE,GL_FALSE,GL_TRUE);
	glDepthFunc(GL_LESS);

        #ifndef NDEBUG
	cout << "Leaving closestSurfaceBehind..." << endl;
        #endif

}

/////////////////////////////////////////////////////
// subtractionEngine
//
//

void 
subtractionEngine(const CsgProduct &product,GLuint layers)
{
	assert( product.size()>0 );
	assert( product.uncomplemented().size()==1 );
	assert( product.complemented().size()>0 );
	assert( layers>0 );

        #ifndef NDEBUG
        cout << "Entering NSD-A..." << endl;
        #endif


	//
	// Detect the depth of the stencil buffer
	//

	GLint		stencilSize = 0;
	glGetIntegerv(GL_STENCIL_BITS,&stencilSize);
	mostSignificantStencilBit = 1<<(stencilSize-1);

	#ifndef NDEBUG
	cout << stencilSize << " bit stencil buffer detected." <<endl;
	#endif

	//
	// Determine the convexity of each object, and store
	// it here.
	//

	vector<unsigned int> convexity;

	for (int n=0; n<product.complemented().size(); n++)
		convexity.push_back(1);

	//
	// Initialise depth & stencil buffers
	//

	glClearDepth(1.0);
	glClearStencil(0x0000);

	glStencilMask(0xffff);
	glDepthMask(GL_TRUE);
	glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);

	glClear(GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	//
	// Depth and Stencil buffers for full screen operation
	//

	GLuint emptyBit     = mostSignificantStencilBit;
	GLuint subtractBit  = 0x0001;
	GLuint subtractedBit = 0x0002;

	#ifndef NDEBUG
	cout << "Initialising buffers..." << endl;
	#endif

	depthBuffer   workPieceDepth(NULL),frontDepth(NULL),backDepth(NULL);
	stencilBuffer workPieceStencil(NULL),frontStencil,backStencil;

	//
	//	Initialise stencil with Empty Flag
	//
	
	glDisable(GL_LIGHTING);

	glDepthColorStencilMasks(GL_FALSE,GL_FALSE,0xffff);
	glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_FALSE);
	glDepthFunc(GL_ALWAYS);
	glStencilFunc(GL_ALWAYS,emptyBit,0xffff);
	glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);
	drawZnear();

	//
	//	Initialise Work-Piece buffer
	//
	//	Stencil & Empty <- ~Empty 
	//

	glDepthColorStencilMasks(GL_TRUE,GL_FALSE,emptyBit);  
	glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_TRUE);
	glDepthFunc(GL_LESS);
	glStencilFunc(GL_ALWAYS,0x0000,0x0000);
	glStencilOp(GL_ZERO,GL_ZERO,GL_ZERO);
	glCullFace(GL_BACK);
	product.uncomplemented()[0]->draw();

	workPieceDepth.read();
	workPieceStencil.read();

	//
	//	Initialise subtraction buffers
	//

	glClearDepth(0.0);
	glDepthMask(GL_TRUE);

	glClear(GL_DEPTH_BUFFER_BIT);

	frontDepth.read();
	backDepth.read();

	//
	//	Subtract convex layers from workpiece.
	//

	for (GLuint k=0;k<layers;k++)
	{

		#ifndef DEBUG
		cout << "Subtracting layer " << k << endl;
		#endif

		//
		// Update back depth buffer
		//

		glDepthColorStencilMasks(GL_TRUE,GL_FALSE,0xffff);
		glDepthStencilCullingEnable(GL_TRUE,GL_FALSE,GL_FALSE);
		glDepthFunc(GL_ALWAYS);

		backDepth.write();
		backStencil.write();

		glEnable(GL_CULL_FACE);
		glCullFace(GL_FRONT);

		#ifndef NDEBUG
		cout << "  Closest back surface behind..." << endl;
		#endif

		GLERROR;
		closestSurfaceBehind(product,convexity);
		GLERROR;

		backDepth.read();
		backStencil.read();

		//
		//	Update front depth buffer
		//

		glDepthColorStencilMasks(GL_TRUE,GL_FALSE,0xffff);
		glDepthStencilCullingEnable(GL_TRUE,GL_FALSE,GL_FALSE);
		glDepthFunc(GL_ALWAYS);

		frontDepth.write();
		frontStencil.write();

		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);

		#ifndef NDEBUG
		cout << "  Closest front surface behind..." << endl;
		#endif

		GLERROR;
		closestSurfaceBehind(product,convexity);
		GLERROR;
		
		frontDepth.read();
		frontStencil.read();

		//
		// Prepare to compare
		//
		
		glDepthColorStencilMasks(GL_TRUE,GL_FALSE,0xffff);
		glDepthStencilCullingEnable(GL_TRUE,GL_FALSE,GL_FALSE);
		glDepthFunc(GL_ALWAYS);

		GLERROR;
		workPieceDepth.write();
		GLERROR;

/*
		{
			static char *filename = "wp0.ppm"; 
			if (filename[2]<='9')
			{
				workPieceDepth.writePPM(filename);
				filename[2]++;
			}
		}
		
               {
                        static char *filename = "bf0.ppm";
                        if (filename[2]<='9')
                        {
                                frontDepth.writePPM(filename);
                                filename[2]++;
                        }
                }

               {
                        static char *filename = "bb0.ppm";
                        if (filename[2]<='9')
                        {
                                backDepth.writePPM(filename);
                                filename[2]++;
                        }
                }
*/	
		
		//
		// Test that front buffer <= tool
		//

		GLERROR;
		glClear(GL_STENCIL_BUFFER_BIT);
		GLERROR;

                { stencilBuffer stencil; stencil.writePPM("before.ppm"); }
		{ depthBuffer depth; depth.writePPM("befored.ppm"); }

		glDepthColorStencilMasks(GL_FALSE,GL_FALSE,0xff);
		glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_FALSE);
		glDepthFunc(GL_LEQUAL);
		
		glStencilFunc(GL_ALWAYS,subtractBit,0xff);
		glStencilOp(GL_ZERO,GL_ZERO,GL_REPLACE);

 		GLERROR;
		frontDepth.write();
		GLERROR;

		{ stencilBuffer stencil; stencil.writePPM("after.ppm"); }

		//
		// Test that back buffer >= tool
		//

		glDepthColorStencilMasks(GL_TRUE,GL_FALSE,0xff);
		glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_FALSE);
		glDepthFunc(GL_NEVER);
		glStencilFunc(GL_EQUAL,subtractBit,0xff);
		glStencilOp(GL_ZERO,GL_ZERO,GL_ZERO);

		glDisable(GL_STENCIL_TEST);
		glDisable(GL_DEPTH_TEST);
//		glDepthMask(GL_FALSE);

		GLERROR;
		backDepth.write();
		GLERROR;


                { depthBuffer depth; depth.writePPM("afterd.ppm"); }

               // Ensure work-piece is in depth buffer

              {
                        workPieceDepth.read();

                        static char *filename = "zb0.ppm";
                        if (filename[2]<='9')
                        {
                                workPieceDepth.writePPM(filename);
                                filename[2]++;
                        }
                }


		//
		// Set flag where subtraction
		// 
/*
		glDepthColorStencilMasks(GL_FALSE,GL_FALSE,subtractedBit);
		glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_FALSE);
		glDepthFunc(GL_ALWAYS);
		glStencilFunc(GL_EQUAL,subtractedBit|subtractBit,subtractBit);
		glStencilOp(GL_KEEP,GL_REPLACE,GL_REPLACE);
		drawZnear();
*/
		//
		//

		GLERROR;

//		if (k+1 != layers)
		{
			workPieceDepth.read();
			workPieceStencil.read();
		}
	}

               {
                        static char *filename = "tl0.ppm";
                        if (filename[2]<='9')
                        {
                                workPieceDepth.writePPM(filename);
                                filename[2]++;
                        }
                }


	//
	// Clear stencil buffer
	//

	GLERROR;

	glStencilMask(0xffff);
	glClearStencil(0x0000);
	glClear(GL_STENCIL_BUFFER_BIT);

	//
	//	Set EMPTY flag where toolDepth > blank
	//

	glDepthColorStencilMasks(GL_FALSE,GL_FALSE,emptyBit);
	glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_TRUE);
	glDepthFunc(GL_LESS);	
	glStencilFunc(GL_ALWAYS,emptyBit,emptyBit);
	glStencilOp(GL_KEEP,GL_KEEP,GL_REPLACE);
	glCullFace(GL_FRONT);
	product.uncomplemented()[0]->draw();

	GLERROR;

	//
	// Set depth to zFar in empty regions
	//

	glDepthColorStencilMasks(GL_TRUE,GL_FALSE,0x0000);
	glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_FALSE);
	glDepthFunc(GL_ALWAYS);
	glStencilFunc(GL_EQUAL,emptyBit,emptyBit);
	glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
	drawZfar();

	//
	// Draw front of blank where not empty and not subtraction
	//

#if 0
	glDepthColorStencilMasks(GL_TRUE,GL_TRUE,0x0000);
	glDepthStencilCullingEnable(GL_TRUE,GL_TRUE,GL_TRUE);
	glDepthFunc(GL_ALWAYS);
	glStencilFunc(GL_EQUAL,0x0000,emptyBit | subtractedBit);
	glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
#else
	glDepthColorStencilMasks(GL_FALSE,GL_TRUE,0x0000);
	glDepthStencilCullingEnable(GL_TRUE,GL_FALSE,GL_TRUE);
	glDepthFunc(GL_EQUAL);
#endif

	glCullFace(GL_BACK);
	glEnable(GL_LIGHTING);

	GLERROR;

	product.uncomplemented()[0]->draw();

	GLERROR;
	workPieceStencil.read();

	//	Draw Subtracted regions into frame buffer.
	//  We match all back facing fragments with Z buffer.

#if 0
	workPieceDepth.read();

	/*	glPixelTransferf(GL_DEPTH_BIAS,-depthBias); */

	glDepthColorStencilMasks(GL_TRUE,GL_FALSE,0xffff);
	glDepthStencilCullingEnable(GL_TRUE,GL_FALSE,GL_FALSE);
	glDepthFunc(GL_ALWAYS);
	workPieceDepth.write();
#endif

	/*	glPixelTransferf(GL_DEPTH_BIAS,0.0); */

	glDepthColorStencilMasks(GL_FALSE,GL_TRUE,0xffff);
	glDepthStencilCullingEnable(GL_TRUE,GL_FALSE,GL_TRUE);
	glCullFace(GL_FRONT);
	glDepthFunc(GL_EQUAL);

	for (int o=0; o<product.complemented().size(); o++)
		product.complemented()[o]->draw();

	#if 0
//	draw = 1;
	closestSurfaceBehind(product,convexity);
//	draw = 0;
	#endif

	//
	// Leave GL state machine in sane state
	//

	glDepthColorStencilMasks(GL_TRUE,GL_TRUE,0xffff);
	glDepthStencilCullingEnable(GL_TRUE,GL_FALSE,GL_FALSE);
	glCullFace(GL_BACK);
	glDepthFunc(GL_LESS);
	glEnable(GL_LIGHTING);

        #ifndef NDEBUG
        cout << "Leaving subtractionEngine..." << endl;
        #endif


}








