//
// OpenGL Solid Modelling
//
// Nigel Stewart (nigels@eisa.net.au)
// RMIT Department of Mechanical and Manufacturing Engineering
// 11th December 1997
//

#ifndef VECTOR3_H
#define VECTOR3_H

#include <atypes.h>
#include <iostream.h>

class Vector
{
public:
	Vector();
	Vector(double a, double b, double c);

	double		 &operator[](int n);
	const double &operator[](int n) const;

	      double &x();
	const double &x() const;
	      double &y();
	const double &y() const;
	      double &z();
	const double &z() const;

	bool operator==(const Vector &v);
	bool operator==(const double &x);

	double operator*(const Vector &v) const;

	Vector &operator= (const Vector &v);
	Vector &operator+=(const Vector &v);
	Vector &operator-=(const Vector &v);

	Vector operator-() const;

	friend Vector operator*(const Vector  &v, const double x);
	friend Vector operator/(const Vector  &v, const double x);
	friend Vector operator*(const double   x, const Vector &v);
	friend Vector operator+(const Vector &v1, const Vector &v2);
	friend Vector operator-(const Vector &v1, const Vector &v2);
	friend Vector xProduct (const Vector &v1, const Vector &v2);

	void normalise();

	double norm() const;
	double dist(const Vector &x) const;

	void glVertex() const;
	void glNormal() const;
	void glColor() const;

	operator double *();

	friend ostream &operator<<(ostream &os, const Vector &x);
	ostream &writePOV(ostream &os) const;

private:
	double _vector[3];
};

const Vector VectorX(1.0,0.0,0.0);
const Vector VectorY(0.0,1.0,0.0);
const Vector VectorZ(0.0,0.0,1.0);
const Vector Vector0(0.0,0.0,0.0);

#endif 
